@@ -1,7 +1,7 @@
-# Copyright (C) 2009-2010 David Sauve, Trapeze.  All rights reserved.
+# Copyright (C) 2009 David Sauve, Trapeze
 
 __author__ = 'David Sauve'
-__version__ = (1, 0, 2, 'beta')
+__version__ = (1, 1, 0, 'beta')
 
 import time
 import datetime
@@ -17,10 +17,10 @@ from django.core.exceptions import ImproperlyConfigured
 from django.utils.encoding import smart_unicode, force_unicode
 
 from haystack.backends import BaseSearchBackend, BaseSearchQuery, SearchNode, log_query
-from haystack.exceptions import HaystackError, MissingDependency
+from haystack.exceptions import MissingDependency
 from haystack.fields import DateField, DateTimeField, IntegerField, FloatField, BooleanField, MultiValueField
 from haystack.models import SearchResult
-from haystack.utils import get_identifier, get_facet_field_name
+from haystack.utils import get_identifier
 
 try:
     import xapian
@@ -32,17 +32,10 @@ DOCUMENT_ID_TERM_PREFIX = 'Q'
 DOCUMENT_CUSTOM_TERM_PREFIX = 'X'
 DOCUMENT_CT_TERM_PREFIX = DOCUMENT_CUSTOM_TERM_PREFIX + 'CONTENTTYPE'
 
-BACKEND_NAME = 'xapian'
-
-
-class InvalidIndexError(HaystackError):
-    """Raised when an index can not be opened."""
-    pass
-
 
 class XHValueRangeProcessor(xapian.ValueRangeProcessor):
-    def __init__(self, backend):
-        self.backend = backend or SearchBackend()
+    def __init__(self, sb):
+        self.sb = sb
         xapian.ValueRangeProcessor.__init__(self)
     
     def __call__(self, begin, end):
@@ -57,7 +50,7 @@ class XHValueRangeProcessor(xapian.ValueRangeProcessor):
         colon = begin.find(':')
         field_name = begin[:colon]
         begin = begin[colon + 1:len(begin)]
-        for field_dict in self.backend.schema:
+        for field_dict in self.sb.schema:
             if field_dict['field_name'] == field_name:
                 if not begin:
                     if field_dict['type'] == 'text':
@@ -134,24 +127,7 @@ class SearchBackend(BaseSearchBackend):
         if not os.path.exists(settings.HAYSTACK_XAPIAN_PATH):
             os.makedirs(settings.HAYSTACK_XAPIAN_PATH)
         
-        if not os.access(settings.HAYSTACK_XAPIAN_PATH, os.W_OK):
-            raise IOError("The path to your Xapian index '%s' is not writable for the current user/group." % settings.HAYSTACK_XAPIAN_PATH)
-        
         self.language = language
-        self._schema = None
-        self._content_field_name = None
-        
-    @property
-    def schema(self):
-        if not self._schema:
-            self._content_field_name, self._schema = self.build_schema(self.site.all_searchfields())            
-        return self._schema
-
-    @property
-    def content_field_name(self):
-        if not self._content_field_name:
-            self._content_field_name, self._schema = self.build_schema(self.site.all_searchfields())            
-        return self._content_field_name
     
     def update(self, index, iterable):
         """
@@ -163,11 +139,10 @@ class SearchBackend(BaseSearchBackend):
             `iterable` -- An iterable of model instances to index
         
         For each object in `iterable`, a document is created containing all
-        of the terms extracted from `index.full_prepare(obj)` with field prefixes, 
-        and 'as-is' as needed.  Also, if the field type is 'text' it will be 
-        stemmed and stored with the 'Z' prefix as well.
+        of the terms extracted from `index.prepare(obj)` with stemming prefixes,
+        field prefixes, and 'as-is'.
         
-        eg. `content:Testing` ==> `testing, Ztest, ZXCONTENTtest, XCONTENTtest`
+        eg. `content:Testing` ==> `testing, Ztest, ZXCONTENTtest`
         
         Each document also contains an extra term in the format:
         
@@ -210,36 +185,9 @@ class SearchBackend(BaseSearchBackend):
                     if field['field_name'] in data.keys():
                         prefix = DOCUMENT_CUSTOM_TERM_PREFIX + field['field_name'].upper()
                         value = data[field['field_name']]
-                        if field['type'] == 'text':
-                            if field['multi_valued'] == 'false':
-                                term = _marshal_term(value)
-                                term_generator.index_text(term)
-                                term_generator.index_text(term, 1, prefix)
-                                if len(term.split()) == 1:
-                                    document.add_term(term)
-                                    document.add_term(prefix + term)
-                                document.add_value(field['column'], _marshal_value(value))
-                            else:
-                                for term in value:
-                                    term = _marshal_term(term)
-                                    term_generator.index_text(term)
-                                    term_generator.index_text(term, 1, prefix)
-                                    if len(term.split()) == 1:
-                                        document.add_term(term)
-                                        document.add_term(prefix + term)
-                        else:
-                            if field['multi_valued'] == 'false':
-                                term = _marshal_term(value)
-                                if len(term.split()) == 1:
-                                    document.add_term(term)
-                                    document.add_term(prefix + term)
-                                    document.add_value(field['column'], _marshal_value(value))
-                            else:
-                                for term in value:
-                                    term = _marshal_term(term)
-                                    if len(term.split()) == 1:
-                                        document.add_term(term)
-                                        document.add_term(prefix + term)
+                        term_generator.index_text(_marshal_term(value))
+                        term_generator.index_text(_marshal_term(value), 1, prefix)
+                        document.add_value(field['column'], _marshal_value(value))
                 
                 document.set_data(pickle.dumps(
                     (obj._meta.app_label, obj._meta.module_name, obj.pk, data),
@@ -286,7 +234,7 @@ class SearchBackend(BaseSearchBackend):
             query = xapian.Query('')
             enquire = xapian.Enquire(database)
             enquire.set_query(query)
-            for match in self._get_enquire_mset(database, enquire, 0, database.get_doccount()):
+            for match in enquire.get_mset(0, database.get_doccount()):
                 database.delete_document(match.docid)
         else:
             for model in models:
@@ -294,14 +242,13 @@ class SearchBackend(BaseSearchBackend):
                     DOCUMENT_CT_TERM_PREFIX + '%s.%s' %
                     (model._meta.app_label, model._meta.module_name)
                 )
-    
     @log_query
     def search(self, query, sort_by=None, start_offset=0, end_offset=None,
                fields='', highlight=False, facets=None, date_facets=None,
                query_facets=None, narrow_queries=None, spelling_query=None,
                limit_to_registered_models=True, **kwargs):
         """
-        Executes the Xapian::query as defined in `query`.
+        Executes the search as defined in `query_string`.
         
         Required arguments:
             `query` -- Search query to execute
@@ -329,7 +276,17 @@ class SearchBackend(BaseSearchBackend):
                     `queries` -- A list of query facets
             If faceting was not used, the `facets` key will not be present
         
-        If `query` is None, returns no results.
+        If `query_string` is empty, returns no results.
+        
+        Otherwise, loads the available fields from the database meta data schema
+        and sets up prefixes for each one along with a prefix for `django_ct`,
+        used to filter by model, and loads the current stemmer instance.
+        
+        Afterwards, executes the Xapian query parser to create a query from
+        `query_string` that is then passed to a new `enquire` instance.
+        
+        The resulting match set is passed to :method:`_process_results` for
+        further processing prior to returning a dictionary with the results.
         
         If `HAYSTACK_INCLUDE_SPELLING` was enabled in `settings.py`, the
         extra flag `FLAG_SPELLING_CORRECTION` will be passed to the query parser
@@ -393,12 +350,12 @@ class SearchBackend(BaseSearchBackend):
         }
         
         if not end_offset:
-            end_offset = database.get_doccount() - start_offset
+            end_offset = database.get_doccount()
         
-        matches = self._get_enquire_mset(database, enquire, start_offset, end_offset)
+        matches = enquire.get_mset(start_offset, (end_offset - start_offset))
         
         for match in matches:
-            app_label, module_name, pk, model_data = pickle.loads(self._get_document_data(database, match.document))
+            app_label, module_name, pk, model_data = pickle.loads(match.document.get_data())
             if highlight:
                 model_data['highlighted'] = {
                     self.content_field_name: self._do_highlight(
@@ -406,7 +363,7 @@ class SearchBackend(BaseSearchBackend):
                     )
                 }
             results.append(
-                SearchResult(app_label, module_name, pk, match.percent, weight=match.weight, **model_data)
+                SearchResult(app_label, module_name, pk, match.weight, **model_data)
             )
         
         if facets:
@@ -467,13 +424,11 @@ class SearchBackend(BaseSearchBackend):
         if not end_offset:
             end_offset = database.get_doccount()
         
-        for match in self._get_enquire_mset(database, enquire, 0, end_offset):
+        for match in enquire.get_mset(0, end_offset):
             rset.add_document(match.docid)
-                
-        query = xapian.Query(
-            xapian.Query.OP_ELITE_SET,
-            [expand.term for expand in enquire.get_eset(match.document.termlist_count(), rset, XHExpandDecider())],
-            match.document.termlist_count()
+        
+        query = xapian.Query(xapian.Query.OP_OR,
+            [expand.term for expand in enquire.get_eset(match.document.termlist_count(), rset, XHExpandDecider())]
         )
         query = xapian.Query(
             xapian.Query.OP_AND_NOT, [query, DOCUMENT_ID_TERM_PREFIX + get_identifier(model_instance)]
@@ -498,14 +453,15 @@ class SearchBackend(BaseSearchBackend):
         enquire.set_query(query)
         
         results = []
-        matches = self._get_enquire_mset(database, enquire, start_offset, end_offset)
+        matches = enquire.get_mset(start_offset, end_offset)
         
         for match in matches:
-            app_label, module_name, pk, model_data = pickle.loads(self._get_document_data(database, match.document))
+            document = match.get_document()
+            app_label, module_name, pk, model_data = pickle.loads(document.get_data())
             results.append(
-                SearchResult(app_label, module_name, pk, match.percent, weight=match.weight, **model_data)
+                SearchResult(app_label, module_name, pk, match.weight, **model_data)
             )
-
+        
         return {
             'results': results,
             'hits': matches.get_matches_estimated(),
@@ -542,7 +498,7 @@ class SearchBackend(BaseSearchBackend):
         qp.set_stemmer(xapian.Stem(self.language))
         qp.set_stemming_strategy(xapian.QueryParser.STEM_SOME)
         qp.add_boolean_prefix('django_ct', DOCUMENT_CT_TERM_PREFIX)
-
+        
         for field_dict in self.schema:
             qp.add_prefix(
                 field_dict['field_name'],
@@ -628,7 +584,7 @@ class SearchBackend(BaseSearchBackend):
         """
         facet_dict = {}
         
-        # DS_TODO: Improve this algorithm.  Currently, runs in O(N^2), ouch.
+        # DS_TODO: Improve this algorithm.  Currently, runs in O(N^3), ouch.
         for field in field_facets:
             facet_list = {}
             
@@ -773,59 +729,28 @@ class SearchBackend(BaseSearchBackend):
     
     def _database(self, writable=False):
         """
-        Private method that returns a xapian.Database for use.
-
+        Private method that returns a xapian.Database for use and sets up
+        schema and content_field definitions.
+        
         Optional arguments:
             ``writable`` -- Open the database in read/write mode (default=False)
-
+        
         Returns an instance of a xapian.Database or xapian.WritableDatabase
         """
         if writable:
+            self.content_field_name, self.schema = self.build_schema(self.site.all_searchfields())
+            
             database = xapian.WritableDatabase(settings.HAYSTACK_XAPIAN_PATH, xapian.DB_CREATE_OR_OPEN)
+            database.set_metadata('schema', pickle.dumps(self.schema, pickle.HIGHEST_PROTOCOL))
+            database.set_metadata('content', pickle.dumps(self.content_field_name, pickle.HIGHEST_PROTOCOL))
         else:
-            try:
-                database = xapian.Database(settings.HAYSTACK_XAPIAN_PATH)
-            except xapian.DatabaseOpeningError:
-                raise InvalidIndexError(u'Unable to open index at %s' % settings.HAYSTACK_XAPIAN_PATH)
-
+            database = xapian.Database(settings.HAYSTACK_XAPIAN_PATH)
+            
+            self.schema = pickle.loads(database.get_metadata('schema'))
+            self.content_field_name = pickle.loads(database.get_metadata('content'))
+        
         return database
-
-    def _get_enquire_mset(self, database, enquire, start_offset, end_offset):
-        """
-        A safer version of Xapian.enquire.get_mset
-
-        Simply wraps the Xapian version and catches any `Xapian.DatabaseModifiedError`,
-        attempting a `database.reopen` as needed.
-
-        Required arguments:
-            `database` -- The database to be read
-            `enquire` -- An instance of an Xapian.enquire object
-            `start_offset` -- The start offset to pass to `enquire.get_mset`
-            `end_offset` -- The end offset to pass to `enquire.get_mset`
-        """
-        try:
-            return enquire.get_mset(start_offset, end_offset)
-        except xapian.DatabaseModifiedError:
-            database.reopen()
-            return enquire.get_mset(start_offset, end_offset)
-
-    def _get_document_data(self, database, document):
-        """
-        A safer version of Xapian.document.get_data
-
-        Simply wraps the Xapian version and catches any `Xapian.DatabaseModifiedError`,
-        attempting a `database.reopen` as needed.
-
-        Required arguments:
-            `database` -- The database to be read
-            `document` -- An instance of an Xapian.document object
-        """
-        try:
-            return document.get_data()
-        except xapian.DatabaseModifiedError:
-            database.reopen()
-            return document.get_data()
-
+    
     def _value_column(self, field):
         """
         Private method that returns the column value slot in the database
@@ -863,19 +788,20 @@ class SearchQuery(BaseSearchQuery):
     It acts as an intermediary between the ``SearchQuerySet`` and the
     ``SearchBackend`` itself.
     """
-    def __init__(self, backend=None, site=None):
+    def __init__(self, backend=None):
         """
         Create a new instance of the SearchQuery setting the backend as
         specified.  If no backend is set, will use the Xapian `SearchBackend`.
         
         Optional arguments:
             ``backend`` -- The ``SearchBackend`` to use (default = None)
-            ``site`` -- The site to use (default = None)
         """
         super(SearchQuery, self).__init__(backend=backend)
-        self.backend = backend or SearchBackend(site=site)
+        self.backend = backend or SearchBackend()
     
     def build_query(self):
+        # DS_TODO: How does stemming work with this new refactor?
+        
         if not self.query_filter:
             query = xapian.Query('')
         else:
@@ -903,8 +829,8 @@ class SearchQuery(BaseSearchQuery):
                 ) for term, value in self.boost.iteritems()
             ]
             query = xapian.Query(
-                xapian.Query.OP_AND_MAYBE, query,
-                xapian.Query(xapian.Query.OP_OR, subqueries)
+                xapian.Query.OP_OR, query,
+                xapian.Query(xapian.Query.OP_AND, subqueries)
             )
         
         return query
@@ -915,7 +841,12 @@ class SearchQuery(BaseSearchQuery):
         for child in search_node.children:
             if isinstance(child, SearchNode):
                 query_list.append(
-                    self._query_from_search_node(child, child.negated)
+                    xapian.Query(
+                        xapian.Query.OP_AND,
+                        self._query_from_search_node(
+                            child, child.negated
+                        )
+                    )
                 )
             else:
                 expression, term = child
@@ -1017,11 +948,15 @@ class SearchQuery(BaseSearchQuery):
         for term in term_list:
             if ' ' in term:
                 query_list.append(
-                    self._phrase_query(term.split(), field)
+                    xapian.Query(
+                        xapian.Query.OP_OR, self._phrase_query(term.split(), field)
+                    )
                 )
             else:
                 query_list.append(
-                    self._term_query(term, field)
+                    xapian.Query(
+                        xapian.Query.OP_OR, self._term_query(term, field)
+                    )
                 )
         if is_not:
             return xapian.Query(xapian.Query.OP_AND_NOT, self._all_query(), xapian.Query(xapian.Query.OP_OR, query_list))
@@ -1041,13 +976,13 @@ class SearchQuery(BaseSearchQuery):
         Returns:
             A xapian.Query
         """
-        if is_not:
-            return xapian.Query(
-                xapian.Query.OP_AND_NOT,
-                self._all_query(),
-                self.backend.parse_query('%s:%s' % (field, term)),
-            )
-        return self.backend.parse_query('%s:%s' % (field, term))
+        sb = SearchBackend()
+        term_list = set()
+        for t in sb._database().allterms():
+            if t.term.startswith(term.rstrip('*')):
+                term_list.add(t.term)
+        
+        return self._filter_in(list(term_list), field, is_not)
     
     def _filter_gt(self, term, field, is_not):
         return self._filter_lte(term, field, is_not=(is_not != True))
@@ -1105,27 +1040,24 @@ class SearchQuery(BaseSearchQuery):
             A xapian.Query
         """
         stem = xapian.Stem(self.backend.language)
-        
-        if field == 'id':
-            return xapian.Query('%s%s' % (DOCUMENT_ID_TERM_PREFIX, term))
-        elif field == 'django_ct':
-            return xapian.Query('%s%s' % (DOCUMENT_CT_TERM_PREFIX, term))
-        elif field:
-            stemmed = 'Z%s%s%s' % (
-                DOCUMENT_CUSTOM_TERM_PREFIX, field.upper(), stem(term)
-            )
-            unstemmed = '%s%s%s' % (
-                DOCUMENT_CUSTOM_TERM_PREFIX, field.upper(), term
+        if field:
+            return xapian.Query(
+                xapian.Query.OP_OR, 
+                xapian.Query('Z%s%s%s' % (
+                        DOCUMENT_CUSTOM_TERM_PREFIX, field.upper(), stem(term)
+                    )
+                ),
+                xapian.Query('%s%s%s' % (
+                        DOCUMENT_CUSTOM_TERM_PREFIX, field.upper(), term
+                    )
+                )
             )
         else:
-            stemmed = 'Z%s' % stem(term)
-            unstemmed = term
-            
-        return xapian.Query(
-            xapian.Query.OP_OR,
-            xapian.Query(stemmed),
-            xapian.Query(unstemmed)
-        )
+            return xapian.Query(
+                xapian.Query.OP_OR,
+                xapian.Query('Z%s' % term),
+                xapian.Query(term)
+            )
     
     def _phrase_query(self, term_list, field=None):
         """
